#!/usr/bin/env python
#Author: Mark Feineigle
#Create Date: 2018/03/11
#Last Modify: 2018/03/11
# Mastering Machine Learning with Python in Six Steps page 249

from sklearn import datasets
import numpy as np
import pandas as pd
from sklearn.cluster import KMeans
from sklearn.preprocessing import StandardScaler
iris = datasets.load_iris()
# Let's convert to dataframe
iris = pd.DataFrame(data=np.c_[iris['data'],
                    iris['target']],
                    columns=iris['feature_names']+['species'])
# let's remove spaces from column name
iris.columns = iris.columns.str.replace(' ','')
iris.head()
X = iris.ix[:,:3] # independent variables
y = iris.species # dependent variable
sc = StandardScaler()
sc.fit(X)
X = sc.transform(X)
# K Means Cluster
model = KMeans(n_clusters=3, random_state=11)
model.fit(X)
print model.labels_

# since its unsupervised the labels have been assigned
# not in line with the actual labels so let's convert all
# the 1s to 0s and 0s to 1s, the 2's look fine
from sklearn import metrics
import matplotlib.pyplot as plt
iris['pred_species'] = np.choose(model.labels_, [1, 0, 2]).astype(np.int64)
print "Accuracy :", metrics.accuracy_score(iris.species, iris.pred_species)
print "Classification report :", metrics.classification_report(iris.species, iris.pred_species)
# Set the size of the plot
plt.figure(figsize=(10,7))
# Create a colormap
colormap = np.array(['red', 'blue', 'green'])
# Plot Sepal
plt.subplot(2, 2, 1)
plt.scatter(iris['sepallength(cm)'],
            iris['sepalwidth(cm)'],
            c=colormap[iris.species.astype(np.int64)],
            marker='o',
            s=50)
plt.xlabel('sepallength(cm)')
plt.ylabel('sepalwidth(cm)')
plt.title('Sepal (Actual)')
plt.subplot(2, 2, 2)
plt.scatter(iris['sepallength(cm)'],
            iris['sepalwidth(cm)'],
            c=colormap[iris.pred_species],
            marker='o',
            s=50)
plt.xlabel('sepallength(cm)')
plt.ylabel('sepalwidth(cm)')
plt.title('Sepal (Predicted)')
plt.subplot(2, 2, 3)
plt.scatter(iris['petallength(cm)'],
            iris['petalwidth(cm)'],
            c=colormap[iris.species.astype(np.int64)],
            marker='o',
            s=50)
plt.xlabel('petallength(cm)')
plt.ylabel('petalwidth(cm)')
plt.title('Petal (Actual)')
plt.subplot(2, 2, 4)
plt.scatter(iris['petallength(cm)'],
            iris['petalwidth(cm)'],
            c=colormap[iris.pred_species],
            marker='o',
            s=50)
plt.xlabel('petallength(cm)')
plt.ylabel('petalwidth(cm)')
plt.title('Petal (Predicted)')
plt.tight_layout()
plt.show()


# elbow method - page 253
from scipy.spatial.distance import cdist, pdist
from sklearn.cluster import KMeans
K = range(1,10)
KM = [KMeans(n_clusters=k).fit(X) for k in K]
centroids = [k.cluster_centers_ for k in KM]
D_k = [cdist(X, cent, 'euclidean') for cent in centroids]
cIdx = [np.argmin(D,axis=1) for D in D_k]
dist = [np.min(D,axis=1) for D in D_k]
avgWithinSS = [sum(d)/X.shape[0] for d in dist]
# Total with-in sum of square
wcss = [sum(d**2) for d in dist]
tss = sum(pdist(X)**2)/X.shape[0]
bss = tss-wcss
varExplained = bss/tss*100
kIdx = 10-1
##### plot ###
kIdx = 2
# elbow curve
# Set the size of the plot
plt.figure(figsize=(10,4))
plt.subplot(1, 2, 1)
plt.plot(K, avgWithinSS, 'b*-')
plt.plot(K[kIdx],
         avgWithinSS[kIdx],
         marker='o',
         markersize=12,
         markeredgewidth=2,
         markeredgecolor='r',
         markerfacecolor='None')
plt.grid(True)
plt.xlabel('Number of clusters')
plt.ylabel('Average within-cluster sum of squares')
plt.title('Elbow for KMeans clustering')
plt.subplot(1, 2, 2)
plt.plot(K, varExplained, 'b*-')
plt.plot(K[kIdx],
         varExplained[kIdx],
         marker='o',
         markersize=12,
         markeredgewidth=2,
         markeredgecolor='r',
         markerfacecolor='None')
plt.grid(True)
plt.xlabel('Number of clusters')
plt.ylabel('Percentage of variance explained')
plt.title('Elbow for KMeans clustering')
plt.tight_layout()
plt.show()


# silhouette method - page 255
#NOTE silhouette_samples() is an unknown function
#s(i) = b(i) - a(i) / max {a(i), b(i)}, where a(i) is
#average dissimilarity of ith item with other data points
#from same cluster, b(i) lowest average dissimilarity of i
#to other cluster to which i is not a member.
#from sklearn.metrics import silhouette_score
#from matplotlib import cm
#score = []
#for n_clusters in range(2,10):
#  kmeans = KMeans(n_clusters=n_clusters)
#  kmeans.fit(X)
#  labels = kmeans.labels_
#  centroids = kmeans.cluster_centers_
#  score.append(silhouette_score(X, labels,
#metric='euclidean'))
## Set the size of the plot
#plt.figure(figsize=(10,4))
#plt.subplot(1, 2, 1)
#plt.plot(score)
#plt.grid(True)
#plt.ylabel("Silouette Score")
#plt.xlabel("k")
#plt.title("Silouette for K-means")
## Initialize the clusterer with n_clusters value and a random generator
#model = KMeans(n_clusters=3, init='k-means++', n_init=10, random_state=0)
#model.fit_predict(X)
#cluster_labels = np.unique(model.labels_)
#n_clusters = cluster_labels.shape[0]
## Compute the silhouette scores for each sample
#silhouette_vals = silhouette_samples(X, model.labels_) #NOTE another undefined function!
#plt.subplot(1, 2, 2)
#y_lower, y_upper = 0, 0
#yticks = []
#for i, c in enumerate(cluster_labels):
#  c_silhouette_vals = silhouette_vals[cluster_labels == c]
#  c_silhouette_vals.sort()
#  y_upper += len(c_silhouette_vals)
#  color = cm.spectral(float(i) / n_clusters)
#  plt.barh(range(y_lower, y_upper), c_silhouette_vals, facecolor=color, edgecolor=color, alpha=0.7)
#  yticks.append((y_lower + y_upper) / 2)
#  y_lower += len(c_silhouette_vals)
#silhouette_avg = np.mean(silhouette_vals)
#plt.yticks(yticks, cluster_labels+1)
## The vertical line for average silhouette score of all the values
#plt.axvline(x=silhouette_avg, color="red", linestyle="--")
#plt.ylabel('Cluster')
#plt.xlabel('Silhouette coefficient')
#plt.title("Silouette for K-means")
#plt.show()

# page 285
# Agglomerative Cluster
from sklearn.cluster import AgglomerativeClustering
model = AgglomerativeClustering(n_clusters=3)
model.fit(X)
iris['pred_species'] = model.labels_
print "Accuracy :", metrics.accuracy_score(iris.species, iris.pred_species)
print "Classification report :", metrics.classification_report(iris.species, iris.pred_species)

# page 259
# dendrogram
from scipy.cluster.hierarchy import cophenet, dendrogram, linkage
from scipy.spatial.distance import pdist
# generate the linkage matrix
Z = linkage(X, 'ward')
c, coph_dists = cophenet(Z, pdist(X))
# calculate full dendrogram
plt.figure(figsize=(25, 10))
plt.title('Agglomerative Hierarchical Clustering Dendrogram')
plt.xlabel('sample index')
plt.ylabel('distance')
dendrogram(Z,
           leaf_rotation=90., # rotates the x axis labels
           leaf_font_size=8., # font size for the x axis labels
          )
plt.tight_layout()
plt.show()
